@@ -1,5 +1,5 @@
 from __future__ import annotations
-from typing import Any, Callable, Iterable, List, Optional, Protocol, Sequence, Set, Tuple, TypeVar, Union
+from typing import Iterable, List, Optional, Sequence, Set, Tuple
 from cp93pytools.methodtools import cached_property
 from .poset_exceptions import (
     PosetExceptions,
@@ -9,22 +9,21 @@ from .help_index import HelpIndex
 from .poset_wbools import WBools
 import pyhash
 import numpy as np
-import numpy.typing as npt
 from collections import deque
 from itertools import product, chain
 from functools import reduce
 import time, sys
 import re
 
-npBoolMatrix = npt.NDArray[np.bool_]
-npUInt64Matrix = npt.NDArray[np.uint64]
+npBoolMatrix = np.ndarray
+npUInt64Matrix = np.ndarray
 
 
 class Relation(HelpIndex, WBools):
-    """
+    '''
     Class for boolean relation matrices intended mostly for asserting that
     a matrix relation can be used with the fully featured Poset class.
-    """
+    '''
 
     def __init__(self, rel: npBoolMatrix):
         shape = tuple(rel.shape)
@@ -73,27 +72,26 @@ class Relation(HelpIndex, WBools):
     @cached_property
     def is_reflexive(self):
         rel = self.rel
-        I = np.where(~rel[np.diag_indices_from(rel)])
-        why = len(I) and f'Not reflexive: rel[{I[0]},{I[0]}] is False'
-        return self._wbool(not len(I), why)
+        I, = np.where(~rel[np.diag_indices_from(rel)])
+        why = I.size and f'Not reflexive: rel[{I[0]},{I[0]}] is False'
+        return self._wbool(not why, why)
 
     @cached_property
     def is_antisymmetric(self):
         rel = self.rel
         eye = np.identity(self.n, dtype=np.bool_)
-        I = np.where(rel & rel.T & ~eye)
-        why = len(I) and f'Not antisymmetric: cycle {I[0]}<={I[1]}<={I[0]}'
-        return self._wbool(not len(I), why)
+        I, J = np.where(rel & rel.T & ~eye)
+        why = I.size and f'Not antisymmetric: cycle {I[0]}<={I[1]}<={I[0]}'
+        return self._wbool(not why, why)
 
     @cached_property
     def is_transitive(self):
         rel = self.rel
         rel2 = np.matmul(rel, rel)
         I, J = np.where(((~rel) & rel2))
-        why = len(
-            I
-        ) and f'Not transitive: rel[{I[0]},{J[0]}] is False but there is a path'
-        return self._wbool(not len(I), why)
+        why = I.size and (
+            f'Not transitive: rel[{I[0]},{J[0]}] is False but there is a path')
+        return self._wbool(not why, why)
 
     @classmethod
     def validate(cls, rel: npBoolMatrix, expect_poset: bool = False):
@@ -113,9 +111,9 @@ class Relation(HelpIndex, WBools):
         G = [[] for _ in range(n)]
         for i in range(n):
             for j in range(n):
-                if rel[i, j]:
+                if rel[i, j] and i != j:
                     G[i].append(j)
-        return tarjan(G)
+        return Tarjan(G).tarjan()
 
     def transitive_closure(self):
         if self.is_transitive:
@@ -126,7 +124,11 @@ class Relation(HelpIndex, WBools):
         return self.__class__(rel)
 
     def transitive_reduction(self, _assume_poset=False):
-        'Compute the transitive reduction of the given relation'
+        ''''
+        Compute the transitive reduction of the given relation
+        Raises an exception if the relation is not a poset
+        The output relation is also known as "Hasse diagram"
+        '''
         if not _assume_poset:
             self.is_poset.assert_explain()
         lt = self.rel.copy()
@@ -138,7 +140,7 @@ class Relation(HelpIndex, WBools):
 
 
 class Poset(HelpIndex, WBools):
-    """
+    '''
     Hashable object that represents an inmutable finite partial order.
     Uses a matrix and hashing is invariant under permutations.
 
@@ -147,7 +149,7 @@ class Poset(HelpIndex, WBools):
         - leq: read only less-or-equal boolean nxn matrix:
             leq[i,j]==True iff i <= j
         - labels: tuple of n strings. Only used for displaying
-    """
+    '''
 
     def __init__(self, leq: npBoolMatrix, labels: Sequence[str] = None,
                  _validate=False):
@@ -204,11 +206,15 @@ class Poset(HelpIndex, WBools):
 
     def show(self, f=None, method='auto', labels=None, save=None):
         '''
-        Use graphviz to display or save self (or the endomorphism f if given)
-        method only affects visualization of f. Can be
-          - auto, labels, arrows, labels_bottom, arrows_bottom.
-        The suffix _bottom hides the visualization of f[i] when f[i]=bottom.
-        Default will enable _bottom if self is a lattice and f preserves lub.
+        Use graphviz to display or save self as a Hasse diagram.
+        The argument "method" (string) only affects visualization
+        of the endomorphism f (if given). It can be
+          - arrows: blue arrow from each node i to f[i]
+          - labels: replace the label i of each node with f[i]
+          - labels_bottom: (no label at i if f[i]=bottom)
+          - arrows_bottom: (no arrow at i if f[i]=bottom)
+          - auto: 'arrows_bottom' if self is a lattice and f preserves lub. 'arrows' otherwise.
+        Hidding bottom is only allowed if self.bottom makes sense.
         '''
 
         methods = ('auto', 'labels', 'arrows', 'labels_bottom', 'arrows_bottom')
@@ -325,7 +331,7 @@ class Poset(HelpIndex, WBools):
             for j in range(m):
                 leq[i, j] = f_leq(elems[i], elems[j])
         leq.flags.writeable = False
-        return cls(leq, labels)
+        return cls(leq, labels, _validate=True)
 
     @cached_property
     def heights(self):
@@ -352,10 +358,19 @@ class Poset(HelpIndex, WBools):
         Graph structure methods
     '''
 
-    def subgraph(self, domain):
+    def _parse_domain(self, domain: List[int] | List[bool]) -> List[int]:
         n = self.n
+        assert len(domain) <= n, f'Invalid domain: {domain}'
+        if len(domain) == n > 0:
+            if isinstance(domain[0], bool):
+                domain = [i for i in range(n) if domain[i]]
+        else:
+            assert len(set(domain)) == len(domain), f'Invalid domain: {domain}'
+        return domain  # type:ignore
+
+    def subgraph(self, domain: List[int] | List[bool]):
+        domain = self._parse_domain(domain)
         m = len(domain)
-        assert len(set(domain)) == m <= n, f'Invalid domain: {domain}'
         leq = self.leq
         sub = np.zeros((m, m), dtype=bool)
         for i in range(m):
@@ -460,7 +475,7 @@ class Poset(HelpIndex, WBools):
             if not leq[x, glb]:
                 reason = f'Not a lattice: {i} glb {j} => {glb} or {x}'
                 return self._wbool(False, reason)
-        return self._wbool(True)
+        return self._wbool(False, 'Unknown reason')
 
     @cached_property
     def lub(self):
@@ -702,7 +717,7 @@ class Poset(HelpIndex, WBools):
         leq = self.leq
         new_leq = leq + np.matmul(leq[:, i:i + 1], leq[j:j + 1, :])
         new_leq.flags.writeable = False
-        obj = self.__class__(new_leq)
+        obj = self.__class__(new_leq, _validate=False)
         if not assume_poset:
             _ = obj.toposort
         return obj
@@ -1224,7 +1239,8 @@ class Poset(HelpIndex, WBools):
 
     def __invert__(self):
         'flip the poset upside down'
-        return self.__class__.from_children(self.parents, labels=self.labels)
+        cls = self.__class__
+        return cls.from_children(self.parents, labels=self.labels)
 
     def __add__(self, other):
         if isinstance(other, int):
@@ -1256,6 +1272,7 @@ class Poset(HelpIndex, WBools):
 
     def add_poset(self, other):
         'stack other above self and connect all self.tops with all other.bottoms'
+        cls = self.__class__
         n = self.n
         C = [
             *([j for j in Ci] for Ci in self.children),
@@ -1264,10 +1281,11 @@ class Poset(HelpIndex, WBools):
         for i in self.tops:
             for j in other.bottoms:
                 C[j + n].append(i)
-        return self.__class__.from_children(C)
+        return cls.from_children(C)
 
     def mul_poset(self, other):
         'poset standard multiplication'
+        cls = self.__class__
         n = self.n
         m = other.n
         labels = [None] * (n * m)
@@ -1279,19 +1297,21 @@ class Poset(HelpIndex, WBools):
                 for k in other.children[j]:
                     G[i + j * n].append(i + k * n)
                 labels[i + j * n] = f'({self.labels[i]},{other.labels[j]})'
-        return self.__class__.from_children(G, labels=labels)
+        return cls.from_children(G, labels=labels)
 
     def or_poset(self, other):
         'put other at the right of self without connections'
+        cls = self.__class__
         n = self.n
         C = [
             *([j for j in Ci] for Ci in self.children),
             *([j + n for j in Ci] for Ci in other.children),
         ]
-        return self.__class__.from_children(C)
+        return cls.from_children(C)
 
     def and_poset(self, other):
         'stack other above self and put self.tops * other.bottoms inbetween'
+        cls = self.__class__
         n = self.n
         nodes = [
             *((-1, i) for i in self.non_tops),
@@ -1318,13 +1338,14 @@ class Poset(HelpIndex, WBools):
         for i, Ci in C.items():
             for j in Ci:
                 children[f[i]].append(f[j])
-        return self.__class__.from_children(children)
+        return cls.from_children(children)
 
     def add_number(self, n):
         'add self with itself n times'
         assert isinstance(n, int) and n >= 0, f'{n}'
+        cls = self.__class__
         if n == 0:
-            out = self.__class__.total(0)
+            out = cls.total(0)
         else:
             out = self._operation_number(lambda a, b: a + b, n)
         return out
@@ -1332,8 +1353,9 @@ class Poset(HelpIndex, WBools):
     def mul_number(self, n):
         'multiply self with itself n times'
         assert isinstance(n, int) and n >= 0, f'{n}'
+        cls = self.__class__
         if n == 0:
-            out = self.__class__.total(1)
+            out = cls.total(1)
         else:
             out = self._operation_number(lambda a, b: a * b, n)
         return out
@@ -1341,8 +1363,9 @@ class Poset(HelpIndex, WBools):
     def or_number(self, n):
         'OR operation of self with itself n times'
         assert isinstance(n, int) and n >= 0, f'{n}'
+        cls = self.__class__
         if n == 0:
-            out = self.__class__.total(0)
+            out = cls.total(0)
         else:
             out = self._operation_number(lambda a, b: a | b, n)
         return out
@@ -1350,8 +1373,9 @@ class Poset(HelpIndex, WBools):
     def and_number(self, n):
         'AND operation of self with itself n times'
         assert isinstance(n, int) and n >= 0, f'{n}'
+        cls = self.__class__
         if n == 0:
-            out = self.__class__.total(1)
+            out = cls.total(1)
         else:
             out = self._operation_number(lambda a, b: a & b, n)
         return out
@@ -1589,7 +1613,7 @@ class Poset(HelpIndex, WBools):
     '''
 
     def help_verbose(self):
-        return """
+        return '''
         Except for n, leq and labels, all other attributes are
         lazy loaded and usually cached.
         
@@ -1626,7 +1650,7 @@ class Poset(HelpIndex, WBools):
             V.show(f)
             print(f)
         V.meta_O.show()
-        """
+        '''
 
     '''
     @section
@@ -1685,6 +1709,27 @@ class Poset(HelpIndex, WBools):
         middle = ((d == 2) * (bot[:, None] * top[None, :])).sum()
         return 2 * bot_top + (middle if self.n > 2 else 0)
 
+    @classmethod
+    def all_lattices_adding(cls, n: int):
+        E = [pair for i in range(1, n - 1) for pair in [(0, i), (i, n - 1)]]
+        MN_poset = cls.from_up_edges(n, E)
+        num = {MN_poset: 0}
+        V = [MN_poset]
+        G = {0: []}
+        q = deque([MN_poset])
+        while q:
+            P = q.popleft()
+            for Q in P.iter_add_edge():
+                if Q not in num:
+                    num[Q] = len(V)
+                    V.append(Q)
+                    G[num[Q]] = []
+                    q.append(Q)
+                G[num[P]].append(num[Q])
+        E = [(i, j) for i in G for j in G[i]]
+        poset_of_posets = cls.from_up_edges(len(V), E)
+        return V, poset_of_posets
+
 
 def floyd_warshall(adj: npBoolMatrix, infinity: int) -> npUInt64Matrix:
     'Compute all pairs shortest distances using Floyd-Warshall algorithm'
@@ -1799,50 +1844,59 @@ def graphviz(
     return
 
 
-def tarjan(G):
-    n = len(G)
-    ext = [i for i in range(n)][::-1]
-    pa, vis, low = [-1] * n, [-1] * n, [-1] * n
-    stack, in_stack = [], [0] * n
-    u = t = 0
-    sccV, invV = [], [-1] * n
-    dfs = []
-    Q = [G[i][::-1] for i in range(n)]
-    while dfs or ext:
-        if not dfs:  # external loop
-            u = ext.pop()
-            if vis[u] == -1:
-                dfs.append(u)
-        elif vis[u] == -1:  # enter dfs
-            vis[u] = low[u] = t
-            t += 1
-            stack.append(u)
-            in_stack[u] = 1
-        elif Q[u]:  # middle of dfs
-            v = Q[u].pop()
-            dfs.append(u)
-            if vis[v] == -1:
-                pa[v] = u
-                dfs.append(v)
-            elif in_stack[v]:
-                low[u] = min(low[u], vis[v])
-        else:  # exit dfs
-            if pa[u] != -1:
-                low[pa[u]] = min(low[pa[u]], low[u])
-                low[pa[u]] = min(low[pa[u]], low[u])
-            if low[u] == vis[u]:
-                scc = []
-                while stack[-1] != u:
-                    scc.append(stack.pop())
-                scc.append(stack.pop())
-                j = len(sccV)
-                for x in scc:
-                    in_stack[x] = 0
-                    invV[x] = j
-                sccV.append(scc)
-    sccE = set()
-    for i in range(n):
+import sys
+
+sys.setrecursionlimit(10**6)
+
+
+class Tarjan(list):
+    vis: List[int]
+    low: List[int]
+    stack: List[int]
+    in_stack: List[int]
+    sccs: List[List[int]]
+    vis_cnt: int
+
+    def tarjan(self):
+        G = self
+        n = len(G)
+        G.vis, G.low = [-1] * n, [0] * n
+        G.stack, G.in_stack = [], [0] * n
+        G.vis_cnt = 0
+        G.sccs = []
+        for i in range(n):
+            if G.vis[i] == -1:
+                G.dfs(i)
+        invV = [-1] * n
+        for i, scc in enumerate(G.sccs):
+            for j in scc:
+                invV[j] = i
+        sccE = set()
+        for i in range(n):
+            for j in G[i]:
+                sccE.add((invV[i], invV[j]))
+        sccE = list(sccE)
+        return G.sccs, sccE
+
+    def dfs(self, i):
+        G = self
+        G.vis[i] = G.low[i] = G.vis_cnt
+        G.vis_cnt += 1
+        G.stack.append(i)
+        G.in_stack[i] = 1
         for j in G[i]:
-            sccE.add((invV[i], invV[j]))
-    sccE = list(sccE)
-    return sccV, sccE
+            if G.vis[j] == -1:
+                G.dfs(j)
+                G.low[i] = min(G.low[i], G.low[j])
+            elif G.in_stack[j]:
+                G.low[i] = min(G.low[i], G.vis[j])
+        if G.low[i] == G.vis[i]:
+            scc = []
+            s = G.stack
+            while s[-1] != i:
+                scc.append(s.pop())
+            scc.append(s.pop())
+            for x in scc:
+                G.in_stack[x] = 0
+            G.sccs.append(scc)
+        return
